import { __assign, __read } from "tslib";
import { slashPathToFirestoreRef, deleteCollection, isDocPath, } from './firebase-utils';
/**
 * @param baseRef - Base RTDB reference
 * @param options - Options for ref
 * @returns RTDB Reference
 */
function optionsToRtdbRef(baseRef, options) {
    var newRef = baseRef;
    var optionsToAdd = [
        'orderByChild',
        'orderByKey',
        'orderByValue',
        'equalTo',
        'limitToFirst',
        'limitToLast',
        'startAt',
        'endAt',
    ];
    // TODO: Switch to reduce over options keys with baseRef as acc
    optionsToAdd.forEach(function (optionName) {
        if (options && options[optionName]) {
            newRef = newRef[optionName](options[optionName]);
        }
    });
    return newRef;
}
/**
 * @param dataVal - Value of data
 * @param firestoreStatics - Statics from firestore instance
 * @returns Value converted into timestamp object if possible
 */
function convertValueToTimestampOrGeoPointIfPossible(dataVal, firestoreStatics) {
    /* eslint-disable-next-line no-underscore-dangle */
    if ((dataVal === null || dataVal === void 0 ? void 0 : dataVal._methodName) === 'FieldValue.serverTimestamp') {
        return firestoreStatics.FieldValue.serverTimestamp();
    }
    if (typeof (dataVal === null || dataVal === void 0 ? void 0 : dataVal.seconds) === 'number' &&
        typeof (dataVal === null || dataVal === void 0 ? void 0 : dataVal.nanoseconds) === 'number') {
        return new firestoreStatics.Timestamp(dataVal.seconds, dataVal.nanoseconds);
    }
    if (typeof (dataVal === null || dataVal === void 0 ? void 0 : dataVal.latitude) === 'number' &&
        typeof (dataVal === null || dataVal === void 0 ? void 0 : dataVal.longitude) === 'number') {
        return new firestoreStatics.GeoPoint(dataVal.latitude, dataVal.longitude);
    }
    return dataVal;
}
/**
 * @param data - Data to be set in firestore
 * @param firestoreStatics - Statics from Firestore object
 * @returns Data to be set in firestore with timestamp
 */
function getDataWithTimestampsAndGeoPoints(data, firestoreStatics) {
    // Exit if no statics are passed
    if (!firestoreStatics) {
        return data;
    }
    return Object.entries(data).reduce(function (acc, _a) {
        var _b, _c;
        var _d = __read(_a, 2), currKey = _d[0], currData = _d[1];
        // Convert nested timestamp if item is an object
        if (typeof currData === 'object' &&
            currData !== null &&
            !Array.isArray(currData) &&
            /* eslint-disable-next-line no-underscore-dangle */
            !currData._methodName &&
            !currData.seconds &&
            !(currData.latitude && currData.longitude)) {
            return __assign(__assign({}, acc), (_b = {}, _b[currKey] = getDataWithTimestampsAndGeoPoints(currData, firestoreStatics), _b));
        }
        var value = Array.isArray(currData)
            ? currData.map(function (dataItem) {
                return convertValueToTimestampOrGeoPointIfPossible(dataItem, firestoreStatics);
            })
            : convertValueToTimestampOrGeoPointIfPossible(currData, firestoreStatics);
        return __assign(__assign({}, acc), (_c = {}, _c[currKey] = value, _c));
    }, {});
}
/**
 * @param adminInstance - firebase-admin instance
 * @param action - Action to run
 * @param actionPath - Path in RTDB
 * @param options - Query options
 * @param data - Data to pass to action
 * @returns Promise which resolves with results of calling RTDB
 */
export function callRtdb(adminInstance, action, actionPath, options, data) {
    /**
     * @param err - Error to handle
     * @returns Promise which rejects
     */
    function handleError(err) {
        /* eslint-disable no-console */
        console.error("cypress-firebase: Error with RTDB \"" + action + "\" at path \"" + actionPath + "\" :", err);
        /* eslint-enable no-console */
        return Promise.reject(err);
    }
    if (action === 'get') {
        return optionsToRtdbRef(adminInstance.database().ref(actionPath), options)
            .once('value')
            .then(function (snap) { return snap.val(); })
            .catch(handleError);
    }
    if (action === 'push') {
        var pushRef_1 = adminInstance.database().ref(actionPath).push();
        return pushRef_1
            .set(data)
            .then(function () { return pushRef_1.key; })
            .catch(handleError);
    }
    // Delete action
    var actionNameMap = {
        delete: 'remove',
    };
    var cleanedActionName = actionNameMap[action] || action;
    return adminInstance
        .database()
        .ref(actionPath)[cleanedActionName](data)
        .then(function () {
        // Prevents Cypress error with message:
        // "You must return a promise, a value, or null to indicate that the task was handled."
        return null;
    })
        .catch(handleError);
}
/**
 * @param adminInstance - firebase-admin instance
 * @param action - Action to run
 * @param actionPath - Path to collection or document within Firestore
 * @param options - Query options
 * @param data - Data to pass to action
 * @returns Promise which resolves with results of calling Firestore
 */
export function callFirestore(adminInstance, action, actionPath, options, data) {
    /**
     * @param err - Error to handle
     * @returns Promise which rejects
     */
    function handleError(err) {
        /* eslint-disable no-console */
        console.error("cypress-firebase: Error with Firestore \"" + action + "\" at path \"" + actionPath + "\" :", err);
        /* eslint-enable no-console */
        return Promise.reject(err);
    }
    if (action === 'get') {
        return slashPathToFirestoreRef(adminInstance.firestore(), actionPath, options)
            .get()
            .then(function (snap) {
            var _a;
            if (snap && ((_a = snap.docs) === null || _a === void 0 ? void 0 : _a.length) && typeof snap.docs.map === 'function') {
                return snap.docs.map(function (docSnap) { return (__assign(__assign({}, docSnap.data()), { id: docSnap.id })); });
            }
            // Falling back to null in the case of falsey value prevents Cypress error with message:
            // "You must return a promise, a value, or null to indicate that the task was handled."
            return (snap && typeof snap.data === 'function' && snap.data()) || null;
        })
            .catch(handleError);
    }
    if (action === 'delete') {
        // Handle deleting of collections & sub-collections if not a doc path
        var deletePromise = isDocPath(actionPath)
            ? slashPathToFirestoreRef(adminInstance.firestore(), actionPath, options).delete()
            : deleteCollection(adminInstance.firestore(), actionPath, options === null || options === void 0 ? void 0 : options.batchSize);
        return (deletePromise
            // Returning null in the case of falsey value prevents Cypress error with message:
            // "You must return a promise, a value, or null to indicate that the task was handled."
            .then(function () { return null; })
            .catch(handleError));
    }
    if (!data) {
        throw new Error("You must define data to run " + action + " in firestore.");
    }
    var dataToSet = getDataWithTimestampsAndGeoPoints(data, 
    // Use static option if passed (tests), otherwise fallback to statics on adminInstance
    // Tests do not have statics since they are using @firebase/testing
    (options === null || options === void 0 ? void 0 : options.statics) || adminInstance.firestore);
    if (action === 'set') {
        return adminInstance
            .firestore()
            .doc(actionPath)
            .set(dataToSet, (options === null || options === void 0 ? void 0 : options.merge) ? { merge: options === null || options === void 0 ? void 0 : options.merge }
            : undefined)
            .catch(handleError);
    }
    // "update" action
    return slashPathToFirestoreRef(adminInstance.firestore(), actionPath, options)[action](dataToSet)
        .catch(handleError);
}
/**
 * Create a custom token
 * @param adminInstance - Admin SDK instance
 * @param uid - UID of user for which the custom token will be generated
 * @param settings - Settings object
 * @returns Promise which resolves with a custom Firebase Auth token
 */
export function createCustomToken(adminInstance, uid, settings) {
    // Use custom claims or default to { isTesting: true }
    var customClaims = (settings === null || settings === void 0 ? void 0 : settings.customClaims) || { isTesting: true };
    // Create auth token
    return adminInstance.auth().createCustomToken(uid, customClaims);
}
/**
 * Get Firebase Auth user based on UID
 * @param adminInstance - Admin SDK instance
 * @param uid - UID of user for which the custom token will be generated
 * @returns Promise which resolves with a custom Firebase Auth token
 */
export function getAuthUser(adminInstance, uid) {
    return adminInstance.auth().getUser(uid);
}
